#!/usr/bin/env python
import os
import sys
import argparse
import logging
import itertools
import re
import multiprocessing
import subprocess
import pandas as pd
from mputil import lazy_map
from functools import partial

from .mutationhash import mutationhash
from .buffered_reader import buffered_blob
from .worker import _demult_chunk, _writer

# https://stackoverflow.com/a/43922107/6198494
def chunker_list(seq, size):
    return (seq[i::size] for i in range(size))

def demultiplex():
    parser = argparse.ArgumentParser(description='Demultiplexing of fastq files')
    parser.add_argument('--fastq', '-f', help='FASTQ file for demultiplexing.', metavar='input.fastq.gz', type=str)
    parser.add_argument('--samplesheet', '-s', help = 'Samplesheet containing barcodes and samplenames', metavar = 'samplesheet.txt', type=str)
    parser.add_argument('--barcode-regex', '-b', help = 'Regular expression to parse cell barcode (CB) and UMIs (UMI) from read names', metavar = '(.*):(?P<CB>[ATGCN]{11}', default = '(.*):(?P<CB>[ATGCN]{11}', type = str)
    parser.add_argument('--edit-distance', help='Maximum allowed edit distance for barcodes', metavar = '1', type=int, default = 1)
    parser.add_argument('--edit-alphabet', help='The alphabet that is used to created edited barcodes', choices=['N', 'ACGT', 'ACGTN'], default = "ACGTN", type = str, metavar = "ACGTN")
    parser.add_argument('--write-unmatched', help='Write reads with unmatched barcodes into unmatched.fastq.gz', action='store_true')
    parser.add_argument('--barcode-column', help='Name of the column containing barcodes', type=str, default='Barcode', metavar = 'Barcode')
    parser.add_argument('--sample-column', help='Name of the column containing sample names', type=str, default='Sample', metavar = 'Sample')
    parser.add_argument('--column-separator', help='Separator that is used in samplesheet', type=str, default='\t')
    parser.add_argument('--buffer-size', help="Buffer size for the FASTQ reader (in Bytes). Must be large enough to contain the largest entry.", type = int, default = 4000000, metavar = '4000000')
    parser.add_argument('--output', '-o', help = "Output directory to write individual fastq files to.", type = str, metavar = 'fastq', default = None)
    parser.add_argument('--threads', '-t', help='Number of threads to use for multiprocessing.', type=int, metavar='1', default=1)
    parser.add_argument('--writer-threads', '-w', help='Number of threads to use for writing', type=int, metavar='2', default=2)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s 0.4')
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()

    logger = logging.getLogger(__name__)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    sh.setLevel(logging.DEBUG)
    logger.addHandler(sh)

    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)

    logger.debug('Working on {} using {} threads'.format(args.fastq, args.threads))

    #
    # Check output folder
    #
    args.output = os.path.abspath(args.output) if args.output != None else os.path.abspath(os.getcwd())
    if not os.access(args.output, os.W_OK):
        logger.error("Error: Directory {} does not exist or is not writeable.".format(args.output))
        sys.exit(1)

    #
    # Create regular expression for barcode parsing from sequence header
    #
    c_barcode_regex = re.compile(args.barcode_regex)

    # Validate regex for presence of CB group and UMI group
    if 'CB' not in c_barcode_regex.groupindex.keys():
        logger.error('No cell barcode group CB found in barcode regex.')
        sys.exit(1)

    #logger.debug('Set barcode regex to {}'.format(barcode_regex))

    #
    # Create mutationhash in case of barcode whitelist
    #
    samplesheet = pd.read_table(args.samplesheet, sep = args.column_separator, index_col = args.sample_column)
    barcode_dict = samplesheet[args.barcode_column].to_dict()
    barcodes = list(barcode_dict.values())

    logger.debug('Found barcodes:' + ','.join(barcodes))

    logger.debug('Creating mutation hash with edit distance {} for {} barcodes.'.format(args.edit_distance, len(barcodes)))
    mut_hash = mutationhash(strings = barcodes, nedit = args.edit_distance, alphabet = list(args.edit_alphabet), log = logger)

    bufsize = args.buffer_size
    manager = multiprocessing.Manager()
    writer_pool = multiprocessing.Pool(args.writer_threads)

    # Handle writing queues
    if args.write_unmatched:
        queues = {'unmatched': manager.Queue()}
        queue_list = [queues['unmatched']]
        writer_pool.apply_async(_writer, (queues['unmatched'], ['unmatched'], args.output), callback = lambda x: print(x))
    else:
        queues = {}
        queue_list = []

    for chunk in chunker_list(list(barcode_dict.keys()), args.writer_threads-1):
        logger.debug('Creating writer queue for samples {}.'.format(','.join(chunk)))
        q = manager.Queue()
        q_bc_dict = dict((k, barcode_dict[k]) for k in chunk)
        writer_pool.apply_async(_writer, (q, q_bc_dict, args.output), callback = lambda x: print(x))
        queue_list.append(q)
        for bc in q_bc_dict.values():
            queues[bc] = q

    zcat = subprocess.Popen(['zcat', args.fastq], stdout=subprocess.PIPE, bufsize = bufsize)
    with zcat.stdout as fh:
        blob_generator = buffered_blob(fh, bufsize)
        demult_chunk = partial(_demult_chunk, mutationhash = mut_hash, regex = c_barcode_regex, write_unmatched = args.write_unmatched, q = queues)
        data = lazy_map(demult_chunk, blob_generator, n_cpus = args.threads)

    for q in queue_list:
        q.put((None, None))
    logger.debug('Shutting down queues.')

    writer_pool.close()
    writer_pool.join()



CHAR_PLUS = ord(b'+')
CHAR_NEWLINE = ord(b'\n')

def buffered_blob(handle, bufsize):
    backlog = b''
    blob = b''

    while True:
        blob = handle.read(bufsize)
        if blob == b'':
            # Could be end of files
            yield backlog
            break

        if backlog != b'':
            blob = backlog + blob

        i = blob.rfind(b'\n@', 0)

        # Make sure no quality line was found
        if (blob[i-1] == CHAR_PLUS) and (blob[i-2] == CHAR_NEWLINE):
            i = blob.rfind(b'\n@', 0, i-2)

        backlog = blob[i+1:len(blob)]
        yield blob[0:i]



import itertools

# With input from Valentine Svensson<valentine@nxn.se

def mutationhash(strings, nedit, alphabet = ["A", "C", "G", "T"], log = None):
    """
    produce a hash with each key a nedit distance substitution for a set of
    strings. values of the hash is the set of strings the substitution could
    have come from
    """
    maxlen = max([len(string) for string in strings])
    indexes = generate_idx(maxlen, nedit, alphabet = alphabet)
    muthash = dict()
    for string in strings:
        if string not in muthash:
            if log is not None:
                log.debug("Added {} -> {} to mutationhash".format(string, string))
            muthash[string] = {string}
        for x in substitution_set(string, indexes):
            if x not in muthash:
                muthash[x] = {string}
                if log is not None:
                    log.debug("Added {} -> {} to mutationhash ".format(x, string))
            else:
                muthash[x].add(string)
                if log is not None:
                    log.debug("Added {} -> {} to mutationhash ".format(x, string))
    return muthash

def substitution_set(string, indexes):
    """
    for a string, return a set of all possible substitutions
    """
    strlen = len(string)
    return [mutate_string(string, x) for x in indexes if valid_substitution(strlen, x)]

def valid_substitution(strlen, index):
    """
    skip performing substitutions that are outside the bounds of the string
    """
    values = index[0]
    return all([strlen > i for i in values])

def generate_idx(maxlen, nedit, alphabet = ["A", "C", "G", "T"]):
    """
    generate all possible nedit edits of a string. each item has the form
    ((index1, index2), 'A', 'G')  for nedit=2
    index1 will be replaced by 'A', index2 by 'G'

    this covers all edits < nedit as well since some of the specified
    substitutions will not change the base
    """
    indexlists = []
    ALPHABETS = [alphabet for x in range(nedit)]
    return list(itertools.product(itertools.combinations(range(maxlen), nedit),
                                  *ALPHABETS))

def mutate_string(string, tomutate):
    strlist = list(string)
    for i, idx in enumerate(tomutate[0]):
        strlist[idx] = tomutate[i+1]
    return "".join(strlist)



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='pydemult',
      version='0.4',
      description='Streamed and parallel demultiplexing of fastq files in python',
      long_description=readme(),
      url='https://github.com/jenzopr/pydemult',
      author='Jens Preussner',
      author_email='jens.preussner@mpi-bn.mpg.de',
      license='MIT',
      packages=['pydemult'],
      entry_points = {
        'console_scripts': ['pydemult = pydemult.pydemult:demultiplex']
      },
      scripts = [],
      install_requires=[
        'pandas',
        'mputil'
      ],
      classifiers = [
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3'
      ],
      zip_safe=False,
      include_package_data=True)



import gzip
import re
import time
import os
from functools import reduce

def entryfunc(blob):
    entries = blob.split(b'\n')
    # (header, sequence, quality)
    return(zip(entries[::4], entries[1::4], entries[3::4]))

def _demult_chunk(chunk, mutationhash, regex, write_unmatched, q):
    start = time.time()
    bc_chunks = dict()
    for entry in entryfunc(chunk):
        is_unmatched = False
        match = regex.match(entry[0].decode('utf-8'))

        if match is not None:
            try:
                bc_match = match.group('CB')
                origin = mutationhash[bc_match]

                if len(origin) > 1:
                    is_unmatched = True
                else:
                    barcode = list(origin)[0]
                    if barcode in bc_chunks:
                        bc_chunks[barcode].append(entry)
                    else:
                        bc_chunks[barcode] = [entry]
            except KeyError:
                is_unmatched = True

        if (match is None or is_unmatched) and write_unmatched:
            if 'unmatched' not in bc_chunks:
                bc_chunks['unmatched'] = [entry]
            else:
                bc_chunks['unmatched'].append(entry)
    parsing = time.time()
    for barcode in bc_chunks.keys():
        fastq = reduce(lambda x, y: x + '{}\n{}\n+\n{}\n'.format(y[0].decode('utf-8'), y[1].decode('utf-8'), y[2].decode('utf-8')), bc_chunks[barcode], '')
        q[barcode].put((barcode, fastq))
    writing = time.time()
    return((len(chunk), parsing-start, writing-parsing))

def _writer(q, barcodes, prefix = ''):
    count = 0
    handles = dict()
    for (sample, barcode) in barcodes.items():
        handles[barcode] = gzip.open(os.path.join(prefix, sample + '.fastq.gz'), 'wb')

    while 1:
        (bc, fastq) = q.get()
        if bc == None:
            break
        count += 1
        handles[bc].write(fastq.encode('utf-8'))

    for (sample, barcode) in barcodes.items():
        handles[barcode].close()
    return(count)
